/*
 *      SCCS:  @(#)fcopy.c	1.10 (96/11/07)
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1993 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 *
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

#ifndef lint
static char sccsid[] = "@(#)fcopy.c	1.10 (96/11/07) TETware release 3.1";
#endif

/************************************************************************

SCCS:   	@(#)fcopy.c	1.10 96/11/07 TETware release 3.1
NAME:		fcopy.c
PRODUCT:	TETware
AUTHOR:		Denis McConalogue, UniSoft Ltd.
DATE CREATED:	June 1993

DESCRIPTION:
	Routines to support the TET_SAVE_FILES and TET_EXEC_IN_PLACE
	configuration options.

MODIFICATIONS:
	Andrew Dingwall, UniSoft Ltd., November 1993
	removed un-needed #includes, added support for non-posix systems

	Andrew Dingwall, UniSoft Ltd., March 1994
	process every file name in a list of patterns

	Andrew Dingwall, UniSoft Ltd., August 1996
	trap attempt to recursively copy a directory into one of its
	subdirectories

	Geoff Clare, UniSoft Ltd., August 1996
	Missing <unistd.h>.

************************************************************************/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#  include <unistd.h>
#include "dtmac.h"
#include "error.h"
#include "tetdir.h"
#include "dtetlib.h"
#include "tcclib.h"


#ifdef NEEDsrcFile
static char srcFile[] = __FILE__;       /* file name for error reporting */
#endif

/* mask used to extract the permissions bits from st_mode */
#define MODEMASK	(S_IRWXU | S_IRWXG | S_IRWXO)

/* static function declarations */
static int rdcopy  PROTOLIST((char *, char *));


/*
**	Copy a file to a specified file or directory. If the 'from'
**	file is a directory, a recursive copy is done.
**
**	return 0 if successful or -1 on error
**
**	note that functions which call this function expect that
**	errno will be set on error return.
**	error() clears errno so errno must be saved when error() is called
**	and restored afterwards
*/

int tet_fcopy(src, dest)
char *src;
char *dest;
{
	char buf[BUFSIZ];
	char destfile[MAXPATH + 1];
	struct STAT_ST st_src, st_dest;
	int dest_exists, dest_is_directory, nread, rc, save_errno;
	char *basename;
	FILE *ifp, *ofp;
	static char fmt[] = "can't copy from directory %.*s to non-directory";
	char msg[MAXPATH + sizeof fmt];

	/* TRACE3(1, 1, "fcopy(): from <%s> to <%s>", src, dest); */

	/* look up the src file */
	if (STAT(src, &st_src) < 0) {
		save_errno = errno;
		error(errno, "can't stat", src);
		errno = save_errno;
		return(-1);
	}

	/*
	** see if dest exists -
	**	if src is a directory, dest must be a directory also
	**	whether it exists or not
	*/
	dest_exists = (STAT(dest, &st_dest) < 0) ? 0 : 1;
	if (dest_exists) {
		dest_is_directory = S_ISDIR(st_dest.st_mode);
		if (S_ISDIR(st_src.st_mode) && !dest_is_directory) {
			(void) sprintf(msg, fmt, MAXPATH, src);
			error(ENOTDIR, msg, dest);
			errno = ENOTDIR;
			return(-1);
		}
	}
	else
		dest_is_directory = 1;

	/* determine the destination file name if dest is a directory */
	if (dest_is_directory) {
		if ((basename = strrchr(src, '/')) == (char *) 0)
			basename = src;
		else
			basename++;
		if (strlen(dest) + strlen(basename) > (size_t) (MAXPATH - 1)) {
			error(0, "filename length exceeds limit", (char *) 0);
			return (-1);
		}
		if (strcmp(basename, ".") != 0) {
			(void) sprintf(destfile, "%s/%s", dest, basename);
			dest = destfile;
		}
	}


	/*
	** if src is a directory, make destination directories as required and
	** do a recursive copy to dest
	*/
	if (S_ISDIR(st_src.st_mode)) {
		if (tet_mkalldirs(dest) < 0) {
			save_errno = errno;
			error(errno, "can't create directory", dest);
			errno = save_errno;
			return(-1);
		}
		else
			return(rdcopy(src, dest));
	}

	/*
	** here if src is not a directory -
	** ignore files other than regular files
	*/
	if (!S_ISREG(st_src.st_mode)) {
		error(0, src, "not copied (not a plain file)");
		return(0);
	}

	/*
	** here, src and dest are both names of plain files
	** (although dest may not exist yet)
	*/

/*
** macro to test whether src and dest refer to the same file
**
** in WIN95: st_dev is the drive number and st_ino is always zero;
** there are no links, so a (case-insensitive) string comparison of the
** file name component after an initial drive specifier is sufficient
**
** on other systems, we do the traditional device and inode number comparison
*/
#  define SRC_IS_DEST \
	(st_src.st_dev == st_dest.st_dev && st_src.st_ino == st_dest.st_ino)

	/*
	** make sure that we are not about to copy a file onto itself
	**
	** if dest_exists is true, st_dest contains valid data;
	** if dest_is_directory is false, dest refers to the destination file
	** (if true, dest now refers to a file yet to be created)
	*/

	if (dest_exists && !dest_is_directory) {
		if (SRC_IS_DEST) {
			error(0, "can't copy file to itself:", src);
			return (-1);
		}
	}

	/*
	**	at last we are ready to do the copy
	*/

	/* TRACE3(1, 1, "FILE COPY from <%s> to <%s>", src, dest); */

	if ((ifp = fopen(src, "rb")) == (FILE *) 0) {
		save_errno = errno;
		error(errno, "can't open", src);
		errno = save_errno;
		return(-1);
	}

	if ((ofp = fopen(dest, "wb")) == (FILE *) 0) {
		save_errno = errno;
		error(errno, "can't open", dest);
		(void) fclose(ifp);
		errno = save_errno;
		return(-1);
	}

	rc = 0;
	while ((nread = fread(buf, sizeof buf[0], sizeof buf, ifp)) > 0) {
		(void) fwrite(buf, sizeof buf[0], (size_t)nread, ofp);
		if (ferror(ofp)) {
			save_errno = errno;
			error(errno, "write error on", dest);
			errno = save_errno;
			rc = -1;
			break;
		}
	}

	if (ferror(ifp)) {
		save_errno = errno;
		error(errno, "read error on", src);
		errno = save_errno;
		rc = -1;
	}

	(void) fclose(ifp);
	if (fclose(ofp) < 0) {
		save_errno = errno;
		error(errno, "close error on", dest);
		errno = save_errno;
		rc = -1;
	}

	/* finally, fix up the mode of the destination file if necessary */
	if (
		rc == 0 &&
		STAT(dest, &st_dest) == 0 &&
		((st_src.st_mode & MODEMASK) != (st_dest.st_mode & MODEMASK)) &&
		CHMOD(dest, st_src.st_mode & MODEMASK) < 0
	) {
		save_errno = errno;
		error(errno, "warning: can't chmod", dest);
		errno = save_errno;
	}

	return(rc);
}

/*
**	tet_pmatch() - Match a string against a pattern.
**
**	Shell metacharacters (*, ?, [, ]) are allowed in the
**	pattern, and the matching rules are the same as for
**	filename completion.
**
**	Returns 1 if the string matches, 0 otherwise.
*/

#define MASK(c) ((c) & 0177)
#define META(c) ((c)=='?' || (c)=='*' || (c)=='[' || (c)==']')

int tet_pmatch(str, pattern)
register char   *str, *pattern;
{
        register int    schar;
        char	c;

        int	unresolved;
        int	lchar;
        int	notflag;


        schar = MASK(*str); str++;
        switch (c = *pattern++) {

        case '[':
                {
                        unresolved = 0;
                        notflag    = 0;
                        lchar = 077777;
 
                        if  (*pattern == '!') {
                                notflag++;
                                pattern++;
                        }
 
                        while (c = *pattern++) {
				switch (c) {
				case ']': 

					return (unresolved ? tet_pmatch(str, pattern) : 0);

				case '-':

					if (notflag) {
                                                if (lchar > schar || schar > *(pattern++)) {
                                                        unresolved++;
                                                } else {
                                                        return(0);
                                                }
                                        } else {
                                                if (lchar <= schar && schar <= (*pattern++)) {
                                                        unresolved++;
                                                }
                                        }
					break;
	
				default:

                                        if (notflag) {
                                                if (schar != (lchar = MASK(c))) {
                                                        unresolved++;
                                                } else {
                                                        return(0);
                                                }
                                        } else {
                                                if (schar == (lchar = MASK(c))) {
                                                        unresolved++;
                                                }
                                        }
					break;	

                        	}
			}
                        return (0);
                }

        case '?':

                return (schar ? tet_pmatch(str, pattern) : 0);

        case '*':

                if (!*pattern)
                        return (1);

                str--;
                while (*str) {
                        if (tet_pmatch(str++, pattern))
                                return (1);
                }
                return(0);

	case '\\':

		if ((c = *pattern) && (META(c) || c=='\\'))
			pattern++;
		else
			c = '\\';

		if (MASK(c) != schar)
			return (0);

		return (tet_pmatch(str, pattern));

        case 0:

                return (schar==0 ? 1:0);

        default:

                if (MASK(c) != schar)
                        return(0);
        }
        return (tet_pmatch(str, pattern));
}

/*
**
**	rdcopy - recursively copy files from one directory to another
**
*/

static int rdcopy (from, to)
char *from;
char *to;
{
	DIR 	*fd_from;
        struct 	dirent *dp;
        char 	fromname[MAXPATH+1];
	int	errcount=0;
	int save_errno;
	static char fmt[] = "recursive directory copy from %.*s to %.*s";
	char msg[sizeof fmt + (MAXPATH * 2)];

	/* TRACE3(1, 1, "rdcopy(): src = <%s>, dest = <%s>", from, to); */

	/* ensure that the destination is not below the source tree */
#define TO_IS_BELOW_FROM	(!strncmp(from, to, strlen(from)))
	if (TO_IS_BELOW_FROM) {
		(void) sprintf(msg, fmt, MAXPATH, from, MAXPATH, to);
		error(0, msg, "would never return!");
		return(-1);
	}

        if ((fd_from = OPENDIR(from)) == (DIR *) 0) {
		save_errno = errno;
		error(errno, "can't open directory", from);
		errno = save_errno;
                return (-1);
        }

        for (;;) {
                if ((dp = READDIR(fd_from)) == (struct dirent *) 0) {
			save_errno = errno;
                        (void) CLOSEDIR(fd_from);
			errno = save_errno;
                        return (errcount > 0 ? -1:0);
                }
                if (strcmp(dp->d_name, ".")==0 || strcmp(dp->d_name, "..")==0)
                        continue;

                if (strlen(from) + strlen(dp->d_name) > (size_t) (MAXPATH-1)) {
			error(0, "filename length exceeds limit", (char *)0);
			errcount++;
                        continue;
                }
                (void) sprintf(fromname, "%s/%s", from, dp->d_name);

		/* if 'fromname' has been deleted by another process, it
		   is not treated as an error */
		if ((tet_fcopy(fromname, to) != 0) && (errno != ENOENT))
                	errcount++;
        }
}

